@@ -63,6 +63,11 @@
             NULL,                                       \
             "File to write log messages in"             \
      },                                                 \
+    { "-debug",                                         \
+            ARG_INT32,                                  \
+            NULL,                                       \
+            "Verbosity level for debugging messages"    \
+     },                                                 \
      { "-mfclogdir",                                    \
              ARG_STRING,                                \
              NULL,                                      \
@@ -74,35 +79,47 @@
             "Directory to log raw audio files to" }
 
 /** Options defining beam width parameters for tuning the search. */
-#define POCKETSPHINX_BEAM_OPTIONS								\
-{ "-beam",											\
-      ARG_FLOAT64,										\
-      "1e-48",											\
-      "Beam width applied to every frame in Viterbi search (smaller values mean wider beam)" },	\
-{ "-wbeam",											\
-      ARG_FLOAT64,										\
-      "7e-29",											\
-      "Beam width applied to word exits" },							\
-{ "-pbeam",											\
-      ARG_FLOAT64,										\
-      "1e-48",											\
-      "Beam width applied to phone transitions" },						\
-{ "-lpbeam",											\
-      ARG_FLOAT64,										\
-      "1e-40",											\
-      "Beam width applied to last phone in words" },						\
-{ "-lponlybeam",										\
-      ARG_FLOAT64,										\
-      "7e-29",											\
-      "Beam width applied to last phone in single-phone words" },				\
-{ "-fwdflatbeam",										\
-      ARG_FLOAT64,										\
-      "1e-64",											\
-      "Beam width applied to every frame in second-pass flat search" },				\
-{ "-fwdflatwbeam",										\
-      ARG_FLOAT64,										\
-      "7e-29",											\
-      "Beam width applied to word exits in second-pass flat search" }
+#define POCKETSPHINX_BEAM_OPTIONS                                       \
+{ "-beam",                                                              \
+      ARG_FLOAT64,                                                      \
+      "1e-48",                                                          \
+      "Beam width applied to every frame in Viterbi search (smaller values mean wider beam)" }, \
+{ "-wbeam",                                                             \
+      ARG_FLOAT64,                                                      \
+      "7e-29",                                                          \
+      "Beam width applied to word exits" },                             \
+{ "-pbeam",                                                             \
+      ARG_FLOAT64,                                                      \
+      "1e-48",                                                          \
+      "Beam width applied to phone transitions" },                      \
+{ "-lpbeam",                                                            \
+      ARG_FLOAT64,                                                      \
+      "1e-40",                                                          \
+      "Beam width applied to last phone in words" },                    \
+{ "-lponlybeam",                                                        \
+      ARG_FLOAT64,                                                      \
+      "7e-29",                                                          \
+      "Beam width applied to last phone in single-phone words" },       \
+{ "-fwdflatbeam",                                                       \
+      ARG_FLOAT64,                                                      \
+      "1e-64",                                                          \
+      "Beam width applied to every frame in second-pass flat search" }, \
+{ "-fwdflatwbeam",                                                      \
+      ARG_FLOAT64,                                                      \
+      "7e-29",                                                          \
+      "Beam width applied to word exits in second-pass flat search" },  \
+{ "-pl_window",                                                         \
+      ARG_INT32,                                                        \
+      "0",                                                              \
+      "Phoneme lookahead window size, in frames" },                     \
+{ "-pl_beam",                                                           \
+      ARG_FLOAT64,                                                      \
+      "1e-10",                                                          \
+      "Beam width applied to phone loop search for lookahead" },        \
+{ "-pl_pbeam",                                                          \
+      ARG_FLOAT64,                                                      \
+      "1e-5",                                                           \
+      "Beam width applied to phone loop transitions for lookahead" }
 
 /** Options defining other parameters for tuning the search. */
 #define POCKETSPHINX_SEARCH_OPTIONS \
@@ -138,10 +155,6 @@
       ARG_INT32,                                                                                \
       "-1",                                                                                     \
       "Maximum number of active HMMs to maintain at each frame (or -1 for no pruning)" },       \
-{ "-maxhistpf",                                                                                 \
-      ARG_INT32,                                                                                \
-      "100",                                                                                    \
-      "Max no. of histories to maintain at each frame (UNUSED)" },                              \
 { "-fwdflatefwid",                                                                              \
       ARG_INT32,                                                                                \
       "4",                                                                     	                \
@@ -227,7 +240,15 @@
 { "-fillprob",										\
       ARG_FLOAT32,									\
       "1e-8",										\
-      "Filler word transition probability" }
+        "Filler word transition probability" }, \
+{ "-bghist",   \
+      ARG_BOOLEAN, \
+      "no", \
+      "Bigram-mode: If TRUE only one BP entry/frame; else one per LM state" }, \
+{ "-lextreedump", \
+      ARG_INT32, \
+      "0", \
+      "Whether to dump the lextree structure to stderr (for debugging), 1 for Ravi's format, 2 for Dot format, Larger than 2 will be treated as Ravi's format" }
 
 /** Command-line options for dictionaries. */
 #define POCKETSPHINX_DICT_OPTIONS \
@@ -266,6 +287,10 @@
       ARG_STRING,                                                               \
       NULL,                                                                     \
       "Model definition input file" },                                          \
+{ "-senmgau", \
+      ARG_STRING,                                                               \
+      NULL,                                                                     \
+      "Senone to codebook mapping input file (usually not needed)" }, \
 { "-tmat",                                                                      \
       ARG_STRING,                                                               \
       NULL,                                                                     \
@@ -298,10 +323,10 @@
       ARG_STRING,                                                               \
       NULL,                                                                     \
       "Senone dump (compressed mixture weights) input file" },                  \
-{ "-sdmap",                                                                     \
+{ "-mllr",                                                                      \
       ARG_STRING,                                                               \
       NULL,                                                                     \
-      "Subspace distribution cluster map file" },                               \
+      "MLLR transformation to apply to means and variances" },                  \
 { "-mmap",                                                                      \
       ARG_BOOLEAN,                                                              \
       "yes",                                                                    \
@@ -313,7 +338,11 @@
 { "-topn",                                                                      \
       ARG_INT32,                                                                \
       "4",                                                                      \
-      "Number of top Gaussians to use in scoring" },                            \
+      "Maximum number of top Gaussians to use in scoring." },                   \
+{ "-topn_beam",                                                                 \
+      ARG_STRING,                                                               \
+      "0",                                                                     \
+      "Beam width used to determine top-N Gaussians (or a list, per-feature)" },\
 { "-kdtree",                                                                    \
       ARG_STRING,                                                               \
       NULL,                                                                     \
