@@ -35,27 +35,6 @@
  *
  * Author: David Huggins-Daines <dhuggins@cs.cmu.edu>
  */
-/* Originally based on gstsphinxsink.c from gnome-voice-control:
- *
- * Copyright (C) 2007  Nickolay V. Shmyrev  <nshmyrev@yandex.ru>
- *
- * gstsphinxsink.c:
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Library General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Library General Public License for more details.
- *
- * You should have received a copy of the GNU Library General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
- */
 
 #ifdef HAVE_CONFIG_H
 #  include <config.h>
@@ -120,11 +99,6 @@ enum
 /* Default command line. (will go away soon and be constructed using properties) */
 static char *default_argv[] = {
     "gst-pocketsphinx",
-#ifdef MODELDIR
-    "-hmm", MODELDIR "/hmm/wsj1",
-    "-lm", MODELDIR "/lm/wsj/wlist5o.3e-7.vp.tg.lm.DMP",
-    "-dict", MODELDIR "/lm/wsj/wlist5o.dic",
-#endif
     "-samprate", "8000",
     "-cmn", "prior",
     "-nfft", "256",
@@ -757,4 +731,9 @@ GST_PLUGIN_DEFINE(GST_VERSION_MAJOR,
                   "pocketsphinx",
                   "PocketSphinx plugin",
                   plugin_init, VERSION,
-                  "LGPL", "PocketSphinx", "http://pocketsphinx.org/")
+#if (GST_VERSION_MINOR == 10 && GST_VERSION_MICRO < 15) /* Nokia's bogus old GStreamer */
+                  "LGPL",
+#else
+                  "BSD",
+#endif
+                  "PocketSphinx", "http://cmusphinx.sourceforge.net/")
