@@ -88,18 +88,23 @@ static ps_searchfuncs_t fsg_funcs = {
 
 ps_search_t *
 fsg_search_init(cmd_ln_t *config,
-                 acmod_t *acmod,
-                 dict_t *dict)
+                acmod_t *acmod,
+                dict_t *dict,
+                dict2pid_t *d2p)
 {
     fsg_search_t *fsgs;
     char const *path;
 
     fsgs = ckd_calloc(1, sizeof(*fsgs));
-    ps_search_init(ps_search_base(fsgs), &fsg_funcs, config, acmod, dict);
+    ps_search_init(ps_search_base(fsgs), &fsg_funcs, config, acmod, dict, d2p);
 
     /* Initialize HMM context. */
     fsgs->hmmctx = hmm_context_init(bin_mdef_n_emit_state(acmod->mdef),
                                     acmod->tmat->tp, NULL, acmod->mdef->sseq);
+    if (fsgs->hmmctx == NULL) {
+        ps_search_free(ps_search_base(fsgs));
+        return NULL;
+    }
 
     /* Intialize the search history object */
     fsgs->history = fsg_history_init(NULL, dict);
@@ -129,7 +134,7 @@ fsg_search_init(cmd_ln_t *config,
         fsgs->bestpath = TRUE;
 
     /* Acoustic score scale for posterior probabilities. */
-    fsgs->ascale = 1.0f / cmd_ln_float32_r(config, "-ascale");
+    fsgs->ascale = 1.0 / cmd_ln_float32_r(config, "-ascale");
 
     E_INFO("FSG(beam: %d, pbeam: %d, wbeam: %d; wip: %d, pip: %d)\n",
            fsgs->beam_orig, fsgs->pbeam_orig, fsgs->wbeam_orig,
@@ -147,7 +152,9 @@ fsg_search_init(cmd_ln_t *config,
         }
         if (fsg_set_select(fsgs, fsg_model_name(fsg)) == NULL)
             goto error_out;
-        if (fsg_search_reinit(ps_search_base(fsgs)) < 0)
+        if (fsg_search_reinit(ps_search_base(fsgs),
+                              ps_search_dict(fsgs),
+                              ps_search_dict2pid(fsgs)) < 0)
             goto error_out;
     }
     /* Or load a JSGF grammar */
@@ -197,7 +204,9 @@ fsg_search_init(cmd_ln_t *config,
         }
         if (fsg_set_select(fsgs, fsg_model_name(fsg)) == NULL)
             goto error_out;
-        if (fsg_search_reinit(ps_search_base(fsgs)) < 0)
+        if (fsg_search_reinit(ps_search_base(fsgs),
+                              ps_search_dict(fsgs),
+                              ps_search_dict2pid(fsgs)) < 0)
             goto error_out;
     }
 
@@ -232,7 +241,7 @@ fsg_search_free(ps_search_t *search)
 }
 
 int
-fsg_search_reinit(ps_search_t *search)
+fsg_search_reinit(ps_search_t *search, dict_t *dict, dict2pid_t *d2p)
 {
     fsg_search_t *fsgs = (fsg_search_t *)search;
 
@@ -240,13 +249,19 @@ fsg_search_reinit(ps_search_t *search)
     if (fsgs->lextree)
         fsg_lextree_free(fsgs->lextree);
 
+    /* Free old dict2pid, dict */
+    ps_search_base_reinit(search, dict, d2p);
+
+    /* Update the number of words (not used by this module though). */
+    search->n_words = dict_size(dict);
+
     /* Allocate new lextree for the given FSG */
-    fsgs->lextree = fsg_lextree_init(fsgs->fsg, ps_search_dict(fsgs),
+    fsgs->lextree = fsg_lextree_init(fsgs->fsg, dict, d2p,
                                      ps_search_acmod(fsgs)->mdef,
                                      fsgs->hmmctx, fsgs->wip, fsgs->pip);
 
     /* Inform the history module of the new fsg */
-    fsg_history_set_fsg(fsgs->history, fsgs->fsg, ps_search_dict(fsgs));
+    fsg_history_set_fsg(fsgs->history, fsgs->fsg, dict);
 
     return 0;
 }
@@ -277,13 +292,10 @@ fsg_search_add_silences(fsg_search_t *fsgs, fsg_model_t *fsg)
                           cmd_ln_float32_r(ps_search_config(fsgs), "-silprob"));
     n_sil = 0;
     /* Add self-loops for all other fillers. */
-    for (wid = dict_to_id(dict, "<sil>") + 1; wid < dict_n_words(dict); ++wid) {
-        char const *word = dict_word_str(dict, wid);
-        /* FIXME: Shouldn't happen?  Also we need a better way to mark fillers. */
-        if (0 == strcmp(word, "<s>") || 0 == strcmp(word, "</s>")) {
-            E_ERROR("WTF, %s=%d > <sil>=%d\n", word, wid, dict_to_id(dict, "<sil>"));
+    for (wid = dict_filler_start(dict); wid < dict_filler_end(dict); ++wid) {
+        char const *word = dict_wordstr(dict, wid);
+        if (wid == dict_startwid(dict) || wid == dict_finishwid(dict))
             continue;
-        }
         fsg_model_add_silence(fsg, word, -1,
                               cmd_ln_float32_r(ps_search_config(fsgs), "-fillprob"));
         ++n_sil;
@@ -292,25 +304,49 @@ fsg_search_add_silences(fsg_search_t *fsgs, fsg_model_t *fsg)
     return n_sil;
 }
 
+/* Scans the dictionary and check if all words are present. */
 static int
-fsg_search_add_altpron(fsg_search_t *fsgs, fsg_model_t *fsg)
+fsg_search_check_dict(fsg_search_t *fsgs, fsg_model_t *fsg)
 {
     dict_t *dict;
-    int n_alt;
     int i;
 
     dict = ps_search_dict(fsgs);
-    /* Scan FSG's vocabulary for words that have alternate pronunciations. */
-    n_alt = 0;
     for (i = 0; i < fsg_model_n_word(fsg); ++i) {
         char const *word;
         int32 wid;
 
         word = fsg_model_word_str(fsg, i);
-        wid = dict_to_id(dict, word);
-        if (wid != NO_WORD) {
-            while ((wid = dict_next_alt(dict, wid)) != NO_WORD) {
-	        fsg_model_add_alt(fsg, word, dict_word_str(dict, wid));
+        wid = dict_wordid(dict, word);
+        if (wid == BAD_S3WID) {
+    	    E_ERROR("The word '%s' is missing in the dictionary\n", word);
+    	    return FALSE;
+    	}
+    }
+
+    return TRUE;
+}
+
+static int
+fsg_search_add_altpron(fsg_search_t *fsgs, fsg_model_t *fsg)
+{
+    dict_t *dict;
+    int n_alt, n_word;
+    int i;
+
+    dict = ps_search_dict(fsgs);
+    /* Scan FSG's vocabulary for words that have alternate pronunciations. */
+    n_alt = 0;
+    n_word = fsg_model_n_word(fsg);
+    for (i = 0; i < n_word; ++i) {
+        char const *word;
+        int32 wid;
+
+        word = fsg_model_word_str(fsg, i);
+        wid = dict_wordid(dict, word);
+        if (wid != BAD_S3WID) {
+            while ((wid = dict_nextalt(dict, wid)) != BAD_S3WID) {
+	        fsg_model_add_alt(fsg, word, dict_wordstr(dict, wid));
     	        ++n_alt;
     	    }
     	}
@@ -335,6 +371,9 @@ fsg_set_add(fsg_search_t *fsgs, char const *name, fsg_model_t *fsg)
     if (name == NULL)
         name = fsg_model_name(fsg);
 
+    if (!fsg_search_check_dict(fsgs, fsg))
+	return NULL;
+
     /* Add silence transitions and alternate words. */
     if (cmd_ln_boolean_r(ps_search_config(fsgs), "-fsgusefiller")
         && !fsg_model_has_sil(fsg))
@@ -495,7 +534,7 @@ fsg_search_hmm_eval(fsg_search_t *fsgs)
         hmm_dump(hmm, stdout);
 #endif
 
-        if (bestscore < score)
+        if (score BETTER_THAN bestscore)
             bestscore = score;
     }
 
@@ -555,7 +594,8 @@ fsg_search_pnode_trans(fsg_search_t *fsgs, fsg_pnode_t * pnode)
          child; child = fsg_pnode_sibling(child)) {
         newscore = hmm_out_score(hmm) + child->logs2prob;
 
-        if ((newscore >= thresh) && (newscore > hmm_in_score(&child->hmm))) {
+        if ((newscore BETTER_THAN thresh)
+            && (newscore BETTER_THAN hmm_in_score(&child->hmm))) {
             /* Incoming score > pruning threshold and > target's existing score */
             if (hmm_frame(&child->hmm) < nf) {
                 /* Child node not yet activated; do so */
@@ -600,9 +640,9 @@ fsg_search_pnode_exit(fsg_search_t *fsgs, fsg_pnode_t * pnode)
      */
     if (fsg_model_is_filler(fsgs->fsg, wid)
         /* FIXME: This might be slow due to repeated calls to dict_to_id(). */
-        || (dict_pronlen(ps_search_dict(fsgs),
-                         dict_to_id(ps_search_dict(fsgs),
-                                    fsg_model_word_str(fsgs->fsg, wid))) == 1)) {
+        || (dict_is_single_phone(ps_search_dict(fsgs),
+                                   dict_wordid(ps_search_dict(fsgs),
+                                               fsg_model_word_str(fsgs->fsg, wid))))) {
         /* Create a dummy context structure that applies to all right contexts */
         fsg_pnode_add_all_ctxt(&ctxt);
 
@@ -781,8 +821,8 @@ fsg_search_word_trans(fsg_search_t *fsgs)
                  */
                 newscore = score + root->logs2prob;
 
-                if ((newscore >= thresh)
-                    && (newscore > hmm_in_score(&root->hmm))) {
+                if ((newscore BETTER_THAN thresh)
+                    && (newscore BETTER_THAN hmm_in_score(&root->hmm))) {
                     if (hmm_frame(&root->hmm) < nf) {
                         /* Newly activated node; add to active list */
                         fsgs->pnode_active_next =
@@ -811,26 +851,20 @@ fsg_search_word_trans(fsg_search_t *fsgs)
 
 
 int
-fsg_search_step(ps_search_t *search)
+fsg_search_step(ps_search_t *search, int frame_idx)
 {
     fsg_search_t *fsgs = (fsg_search_t *)search;
     int16 const *senscr;
-    int frame_idx, best_senid;
-    int16 best_senscr;
     acmod_t *acmod = search->acmod;
     gnode_t *gn;
     fsg_pnode_t *pnode;
     hmm_t *hmm;
 
-    /* Determine if we actually have a frame to process. */
-    if (acmod->n_feat_frame == 0)
-        return 0;
-
     /* Activate our HMMs for the current frame if need be. */
     if (!acmod->compallsen)
         fsg_search_sen_active(fsgs);
     /* Compute GMM scores for the current frame. */
-    senscr = acmod_score(acmod, &frame_idx, &best_senscr, &best_senid);
+    senscr = acmod_score(acmod, &frame_idx);
     fsgs->n_sen_eval += acmod->n_senone_active;
     hmm_context_set_senscore(fsgs->hmmctx, senscr);
 
@@ -1037,7 +1071,7 @@ fsg_search_find_exit(fsg_search_t *fsgs, int frame_idx, int final, int32 *out_sc
         fl = fsg_hist_entry_fsglink(hist_entry);
         score = fsg_hist_entry_score(hist_entry);
 
-        if (score > bestscore) {
+        if (score BETTER_THAN bestscore) {
             /* Only enforce the final state constraint if this is a final hypothesis. */
             if ((!final)
                 || fsg_link_to_state(fl) == fsg_model_final_state(fsg)) {
@@ -1090,6 +1124,7 @@ char const *
 fsg_search_hyp(ps_search_t *search, int32 *out_score)
 {
     fsg_search_t *fsgs = (fsg_search_t *)search;
+    dict_t *dict = ps_search_dict(search);
     char *c;
     size_t len;
     int bp, bpidx;
@@ -1117,31 +1152,44 @@ fsg_search_hyp(ps_search_t *search, int32 *out_score)
     while (bp > 0) {
         fsg_hist_entry_t *hist_entry = fsg_history_entry_get(fsgs->history, bp);
         fsg_link_t *fl = fsg_hist_entry_fsglink(hist_entry);
+        char const *baseword;
         int32 wid;
 
         bp = fsg_hist_entry_pred(hist_entry);
         wid = fsg_link_wid(fl);
         if (wid < 0 || fsg_model_is_filler(fsgs->fsg, wid))
             continue;
-        len += strlen(fsg_model_word_str(fsgs->fsg, wid)) + 1;
+        baseword = dict_basestr(dict,
+                                dict_wordid(dict,
+                                            fsg_model_word_str(fsgs->fsg, wid)));
+        len += strlen(baseword) + 1;
     }
-
+    
     ckd_free(search->hyp_str);
+    if (len == 0) {
+	search->hyp_str = NULL;
+	return search->hyp_str;
+    }
     search->hyp_str = ckd_calloc(1, len);
+
     bp = bpidx;
     c = search->hyp_str + len - 1;
     while (bp > 0) {
         fsg_hist_entry_t *hist_entry = fsg_history_entry_get(fsgs->history, bp);
         fsg_link_t *fl = fsg_hist_entry_fsglink(hist_entry);
+        char const *baseword;
         int32 wid;
 
         bp = fsg_hist_entry_pred(hist_entry);
         wid = fsg_link_wid(fl);
         if (wid < 0 || fsg_model_is_filler(fsgs->fsg, wid))
             continue;
-        len = strlen(fsg_model_word_str(fsgs->fsg, wid));
+        baseword = dict_basestr(dict,
+                                dict_wordid(dict,
+                                            fsg_model_word_str(fsgs->fsg, wid)));
+        len = strlen(baseword);
         c -= len;
-        memcpy(c, fsg_model_word_str(fsgs->fsg, wid), len);
+        memcpy(c, baseword, len);
         if (c > search->hyp_str) {
             --c;
             *c = ' ';
@@ -1302,7 +1350,7 @@ new_node(ps_lattice_t *dag, fsg_model_t *fsg, int sf, int ef, int32 wid, int32 a
         if (node->fef == -1 || node->fef > ef)
             node->fef = ef;
         /* Update best link score. */
-        if (node->info.best_exit < ascr)
+        if (ascr BETTER_THAN node->info.best_exit)
             node->info.best_exit = ascr;
     }
     else {
@@ -1608,9 +1656,9 @@ fsg_search_lattice(ps_search_t *search)
      * Convert word IDs from FSG to dictionary.
      */
     for (node = dag->nodes; node; node = node->next) {
-        node->wid = dict_to_id(dag->search->dict,
-                               fsg_model_word_str(fsg, node->wid));
-        node->basewid = dict_base_wid(dag->search->dict, node->wid);
+        node->wid = dict_wordid(dag->search->dict,
+                                fsg_model_word_str(fsg, node->wid));
+        node->basewid = dict_basewid(dag->search->dict, node->wid);
     }
 
     /*
