@@ -46,6 +46,7 @@
 /* SphinxBase headers. */
 #include <ckd_alloc.h>
 #include <listelem_alloc.h>
+#include <err.h>
 
 /* Local headers. */
 #include "ngram_search.h"
@@ -64,26 +65,83 @@ ngram_fwdflat_expand_all(ngram_search_t *ngs)
 {
     int n_words, i;
 
+    /* For all "real words" (not fillers or <s>/</s>) in the dictionary,
+     *
+     * 1) Add the ones which are in the LM to the fwdflat wordlist
+     * 2) And to the expansion list (since we are expanding all)
+     */
     ngs->n_expand_words = 0;
     n_words = ps_search_n_words(ngs);
     bitvec_clear_all(ngs->expand_word_flag, ps_search_n_words(ngs));
-    for (i = 0; i < ps_search_start_wid(ngs); i++) {
-        if (ngram_model_set_known_wid(ngs->lmset,
-                                      ps_search_dict(ngs)->dict_list[i]->wid)) {
-            ngs->fwdflat_wordlist[ngs->n_expand_words] = i;
-            ngs->expand_word_list[ngs->n_expand_words] = i;
-            bitvec_set(ngs->expand_word_flag, i);
-            ngs->n_expand_words++;
-        }
+    for (i = 0; i < n_words; ++i) {
+        if (!dict_real_word(ps_search_dict(ngs), i))
+            continue;
+        if (!ngram_model_set_known_wid(ngs->lmset,
+                                       dict_basewid(ps_search_dict(ngs),i)))
+            continue;
+        ngs->fwdflat_wordlist[ngs->n_expand_words] = i;
+        ngs->expand_word_list[ngs->n_expand_words] = i;
+        bitvec_set(ngs->expand_word_flag, i);
+        ngs->n_expand_words++;
     }
+    E_INFO("Utterance vocabulary contains %d words\n", ngs->n_expand_words);
     ngs->expand_word_list[ngs->n_expand_words] = -1;
     ngs->fwdflat_wordlist[ngs->n_expand_words] = -1;
 }
 
+static void
+ngram_fwdflat_allocate_1ph(ngram_search_t *ngs)
+{
+    dict_t *dict = ps_search_dict(ngs);
+    int n_words = ps_search_n_words(ngs);
+    int i, w;
+
+    /* Allocate single-phone words, since they won't have
+     * been allocated for us by fwdtree initialization. */
+    ngs->n_1ph_words = 0;
+    for (w = 0; w < n_words; w++) {
+        if (dict_is_single_phone(dict, w))
+            ++ngs->n_1ph_words;
+    }
+    ngs->rhmm_1ph = ckd_calloc(ngs->n_1ph_words, sizeof(*ngs->rhmm_1ph));
+    i = 0;
+    for (w = 0; w < n_words; w++) {
+        if (!dict_is_single_phone(dict, w))
+            continue;
+
+        /* DICT2PID location */
+        ngs->rhmm_1ph[i].ciphone = dict_first_phone(dict, w);
+        ngs->rhmm_1ph[i].ci2phone = bin_mdef_silphone(ps_search_acmod(ngs)->mdef);
+        hmm_init(ngs->hmmctx, &ngs->rhmm_1ph[i].hmm, TRUE,
+                 /* ssid */ bin_mdef_pid2ssid(ps_search_acmod(ngs)->mdef,
+                                              ngs->rhmm_1ph[i].ciphone),
+                 /* tmatid */ ngs->rhmm_1ph[i].ciphone);
+        ngs->rhmm_1ph[i].next = NULL;
+        ngs->word_chan[w] = (chan_t *) &(ngs->rhmm_1ph[i]);
+        i++;
+    }
+}
+
+static void
+ngram_fwdflat_free_1ph(ngram_search_t *ngs)
+{
+    int i, w;
+    int n_words = ps_search_n_words(ngs);
+
+    for (i = w = 0; w < n_words; ++w) {
+        if (!dict_is_single_phone(ps_search_dict(ngs), w))
+            continue;
+        hmm_deinit(&ngs->rhmm_1ph[i].hmm);
+        ++i;
+    }
+    ckd_free(ngs->rhmm_1ph);
+    ngs->rhmm_1ph = NULL;
+}
+
 void
 ngram_fwdflat_init(ngram_search_t *ngs)
 {
-    int n_words, i;
+    int n_words;
 
     n_words = ps_search_n_words(ngs);
     ngs->fwdflat_wordlist = ckd_calloc(n_words + 1, sizeof(*ngs->fwdflat_wordlist));
@@ -97,34 +155,10 @@ ngram_fwdflat_init(ngram_search_t *ngs)
 
     /* No tree-search; pre-build the expansion list, including all LM words. */
     if (!ngs->fwdtree) {
-        int w;
-
         /* Build full expansion list from LM words. */
         ngram_fwdflat_expand_all(ngs);
-
-        /* Allocate single-phone words, since they won't have
-         * been allocated for us by fwdtree initialization. */
-        ngs->n_1ph_words = 0;
-        for (w = 0; w < n_words; w++) {
-            dict_entry_t *de = ps_search_dict(ngs)->dict_list[w];
-            if (de->len == 1)
-                ++ngs->n_1ph_words;
-        }
-        ngs->rhmm_1ph = ckd_calloc(ngs->n_1ph_words, sizeof(*ngs->rhmm_1ph));
-        i = 0;
-        for (w = 0; w < n_words; w++) {
-            dict_entry_t *de = ps_search_dict(ngs)->dict_list[w];
-            if (de->len != 1)
-                continue;
-
-            ngs->rhmm_1ph[i].diphone = de->phone_ids[0];
-            ngs->rhmm_1ph[i].ciphone = de->ci_phone_ids[0];
-            hmm_init(ngs->hmmctx, &ngs->rhmm_1ph[i].hmm, de->mpx,
-                     de->phone_ids[0], de->ci_phone_ids[0]);
-            ngs->rhmm_1ph[i].next = NULL;
-            ngs->word_chan[w] = (chan_t *) &(ngs->rhmm_1ph[i]);
-            i++;
-        }
+        /* Allocate single phone words. */
+        ngram_fwdflat_allocate_1ph(ngs);
     }
 }
 
@@ -133,7 +167,7 @@ ngram_fwdflat_deinit(ngram_search_t *ngs)
 {
     /* Free single-phone words if we allocated them. */
     if (!ngs->fwdtree) {
-        ckd_free(ngs->rhmm_1ph);
+        ngram_fwdflat_free_1ph(ngs);
     }
     ckd_free(ngs->fwdflat_wordlist);
     bitvec_free(ngs->expand_word_flag);
@@ -144,10 +178,29 @@ ngram_fwdflat_deinit(ngram_search_t *ngs)
 int
 ngram_fwdflat_reinit(ngram_search_t *ngs)
 {
-    /* No tree-search; re-build the expansion list from all LM words. */
+    /* Reallocate things that depend on the number of words. */
+    int n_words;
+
+    ckd_free(ngs->fwdflat_wordlist);
+    ckd_free(ngs->expand_word_list);
+    bitvec_free(ngs->expand_word_flag);
+    n_words = ps_search_n_words(ngs);
+    ngs->fwdflat_wordlist = ckd_calloc(n_words + 1, sizeof(*ngs->fwdflat_wordlist));
+    ngs->expand_word_flag = bitvec_alloc(n_words);
+    ngs->expand_word_list = ckd_calloc(n_words + 1, sizeof(*ngs->expand_word_list));
+    
+    /* No tree-search; take care of the expansion list and single phone words. */
     if (!ngs->fwdtree) {
+        /* Free single-phone words. */
+        ngram_fwdflat_free_1ph(ngs);
+        /* Reallocate word_chan. */
+        ckd_free(ngs->word_chan);
+        ngs->word_chan = ckd_calloc(dict_size(ps_search_dict(ngs)),
+                                    sizeof(*ngs->word_chan));
         /* Rebuild full expansion list from LM words. */
         ngram_fwdflat_expand_all(ngs);
+        /* Allocate single phone words. */
+        ngram_fwdflat_allocate_1ph(ngs);
     }
     /* Otherwise there is nothing to do since the wordlist is
      * generated anew every utterance. */
@@ -161,14 +214,16 @@ static void
 build_fwdflat_wordlist(ngram_search_t *ngs)
 {
     int32 i, f, sf, ef, wid, nwd;
+    dict_t *dict;
     bptbl_t *bp;
     ps_latnode_t *node, *prevnode, *nextnode;
-    dict_entry_t *de;
 
     /* No tree-search, use statically allocated wordlist. */
     if (!ngs->fwdtree)
         return;
 
+    dict = ps_search_dict(ngs);
+
     memset(ngs->frm_wordlist, 0, ngs->n_frame_alloc * sizeof(*ngs->frm_wordlist));
 
     /* Scan the backpointer table for all active words and record
@@ -184,11 +239,10 @@ build_fwdflat_wordlist(ngram_search_t *ngs)
          * transition can be made in the LM.
          */
         /* Ignore silence and <s> */
-        if (ISA_FILLER_WORD(ngs, wid) || (wid == ps_search_start_wid(ngs)))
+        if (dict_filler_word(dict, wid) || (wid == dict_startwid(dict)))
             continue;
 
         /* Look for it in the wordlist. */
-        de = ps_search_dict(ngs)->dict_list[wid];
         for (node = ngs->frm_wordlist[sf]; node && (node->wid != wid);
              node = node->next);
 
@@ -238,6 +292,7 @@ build_fwdflat_wordlist(ngram_search_t *ngs)
         }
     }
     ngs->fwdflat_wordlist[nwd] = -1;
+    E_INFO("Utterance vocabulary contains %d words\n", nwd);
 }
 
 /**
@@ -247,38 +302,44 @@ static void
 build_fwdflat_chan(ngram_search_t *ngs)
 {
     int32 i, wid, p;
-    dict_entry_t *de;
     root_chan_t *rhmm;
     chan_t *hmm, *prevhmm;
+    dict_t *dict;
+    dict2pid_t *d2p;
+
+    dict = ps_search_dict(ngs);
+    d2p = ps_search_dict2pid(ngs);
 
     /* Build word HMMs for each word in the lattice. */
     for (i = 0; ngs->fwdflat_wordlist[i] >= 0; i++) {
         wid = ngs->fwdflat_wordlist[i];
-        de = ps_search_dict(ngs)->dict_list[wid];
 
         /* Omit single-phone words as they are permanently allocated */
-        if (de->len == 1)
+        if (dict_is_single_phone(dict, wid))
             continue;
 
-        assert(de->mpx);
         assert(ngs->word_chan[wid] == NULL);
 
         /* Multiplex root HMM for first phone (one root per word, flat
-         * lexicon) */
+         * lexicon).  diphone is irrelevant here, for the time being,
+         * at least. */
         rhmm = listelem_malloc(ngs->root_chan_alloc);
-        rhmm->diphone = de->phone_ids[0];
-        rhmm->ciphone = de->ci_phone_ids[0];
+        rhmm->ci2phone = dict_second_phone(dict, wid);
+        rhmm->ciphone = dict_first_phone(dict, wid);
         rhmm->next = NULL;
-        hmm_init(ngs->hmmctx, &rhmm->hmm, TRUE, rhmm->diphone, rhmm->ciphone);
+        hmm_init(ngs->hmmctx, &rhmm->hmm, TRUE,
+                 bin_mdef_pid2ssid(ps_search_acmod(ngs)->mdef, rhmm->ciphone),
+                 rhmm->ciphone);
 
         /* HMMs for word-internal phones */
         prevhmm = NULL;
-        for (p = 1; p < de->len - 1; p++) {
+        for (p = 1; p < dict_pronlen(dict, wid) - 1; p++) {
             hmm = listelem_malloc(ngs->chan_alloc);
-            hmm->ciphone = de->ci_phone_ids[p];
-            hmm->info.rc_id = p + 1 - de->len;
+            hmm->ciphone = dict_pron(dict, wid, p);
+            hmm->info.rc_id = (p == dict_pronlen(dict, wid) - 1) ? 0 : -1;
             hmm->next = NULL;
-            hmm_init(ngs->hmmctx, &hmm->hmm, FALSE, de->phone_ids[p], hmm->ciphone);
+            hmm_init(ngs->hmmctx, &hmm->hmm, FALSE,
+                     dict2pid_internal(d2p,wid,p), hmm->ciphone);
 
             if (prevhmm)
                 prevhmm->next = hmm;
@@ -382,7 +443,7 @@ fwdflat_eval_chan(ngram_search_t *ngs, int frame_idx)
         rhmm = (root_chan_t *) ngs->word_chan[w];
         if (hmm_frame(&rhmm->hmm) == frame_idx) {
             int32 score = chan_v_eval(rhmm);
-            if ((bestscore < score) && (w != ps_search_finish_wid(ngs)))
+            if ((score BETTER_THAN bestscore) && (w != ps_search_finish_wid(ngs)))
                 bestscore = score;
             ngs->st.n_fwdflat_chan++;
         }
@@ -390,7 +451,7 @@ fwdflat_eval_chan(ngram_search_t *ngs, int frame_idx)
         for (hmm = rhmm->next; hmm; hmm = hmm->next) {
             if (hmm_frame(&hmm->hmm) == frame_idx) {
                 int32 score = chan_v_eval(hmm);
-                if (bestscore < score)
+                if (score BETTER_THAN bestscore)
                     bestscore = score;
                 ngs->st.n_fwdflat_chan++;
             }
@@ -407,7 +468,6 @@ fwdflat_prune_chan(ngram_search_t *ngs, int frame_idx)
     int32 *awl;
     root_chan_t *rhmm;
     chan_t *hmm, *nexthmm;
-    dict_entry_t *de;
 
     cf = frame_idx;
     nf = cf + 1;
@@ -418,31 +478,30 @@ fwdflat_prune_chan(ngram_search_t *ngs, int frame_idx)
     thresh = ngs->best_score + ngs->fwdflatbeam;
     wordthresh = ngs->best_score + ngs->fwdflatwbeam;
     pip = ngs->pip;
+    E_DEBUG(3,("frame %d thresh %d wordthresh %d\n", frame_idx, thresh, wordthresh));
 
     /* Scan all active words. */
     for (w = *(awl++); i > 0; --i, w = *(awl++)) {
-        de = ps_search_dict(ngs)->dict_list[w];
-
         rhmm = (root_chan_t *) ngs->word_chan[w];
         /* Propagate active root channels */
         if (hmm_frame(&rhmm->hmm) == cf
-            && hmm_bestscore(&rhmm->hmm) > thresh) {
+            && hmm_bestscore(&rhmm->hmm) BETTER_THAN thresh) {
             hmm_frame(&rhmm->hmm) = nf;
             bitvec_set(ngs->word_active, w);
 
             /* Transitions out of root channel */
             newscore = hmm_out_score(&rhmm->hmm);
             if (rhmm->next) {
-                assert(de->len > 1);
+                assert(!dict_is_single_phone(ps_search_dict(ngs), w));
 
                 newscore += pip;
-                if (newscore > thresh) {
+                if (newscore BETTER_THAN thresh) {
                     hmm = rhmm->next;
                     /* Enter all right context phones */
                     if (hmm->info.rc_id >= 0) {
                         for (; hmm; hmm = hmm->next) {
                             if ((hmm_frame(&hmm->hmm) < cf)
-                                || (hmm_in_score(&hmm->hmm) < newscore)) {
+                                || (newscore BETTER_THAN hmm_in_score(&hmm->hmm))) {
                                 hmm_enter(&hmm->hmm, newscore,
                                           hmm_out_history(&rhmm->hmm), nf);
                             }
@@ -451,7 +510,7 @@ fwdflat_prune_chan(ngram_search_t *ngs, int frame_idx)
                     /* Just a normal word internal phone */
                     else {
                         if ((hmm_frame(&hmm->hmm) < cf)
-                            || (hmm_in_score(&hmm->hmm) < newscore)) {
+                            || (newscore BETTER_THAN hmm_in_score(&hmm->hmm))) {
                                 hmm_enter(&hmm->hmm, newscore,
                                           hmm_out_history(&rhmm->hmm), nf);
                         }
@@ -459,11 +518,11 @@ fwdflat_prune_chan(ngram_search_t *ngs, int frame_idx)
                 }
             }
             else {
-                assert(de->len == 1);
+                assert(dict_is_single_phone(ps_search_dict(ngs), w));
 
                 /* Word exit for single-phone words (where did their
                  * whmms come from?) */
-                if (newscore > wordthresh) {
+                if (newscore BETTER_THAN wordthresh) {
                     ngram_search_save_bp(ngs, cf, w, newscore,
                                          hmm_out_history(&rhmm->hmm), 0);
                 }
@@ -474,7 +533,7 @@ fwdflat_prune_chan(ngram_search_t *ngs, int frame_idx)
         for (hmm = rhmm->next; hmm; hmm = hmm->next) {
             if (hmm_frame(&hmm->hmm) >= cf) {
                 /* Propagate forward HMMs inside the beam. */
-                if (hmm_bestscore(&hmm->hmm) > thresh) {
+                if (hmm_bestscore(&hmm->hmm) BETTER_THAN thresh) {
                     hmm_frame(&hmm->hmm) = nf;
                     bitvec_set(ngs->word_active, w);
 
@@ -482,14 +541,14 @@ fwdflat_prune_chan(ngram_search_t *ngs, int frame_idx)
                     /* Word-internal phones */
                     if (hmm->info.rc_id < 0) {
                         newscore += pip;
-                        if (newscore > thresh) {
+                        if (newscore BETTER_THAN thresh) {
                             nexthmm = hmm->next;
                             /* Enter all right-context phones. */
                             if (nexthmm->info.rc_id >= 0) {
-                                for (; nexthmm; nexthmm = nexthmm->next) {
+                                 for (; nexthmm; nexthmm = nexthmm->next) {
                                     if ((hmm_frame(&nexthmm->hmm) < cf)
-                                        || (hmm_in_score(&nexthmm->hmm)
-                                            < newscore)) {
+                                        || (newscore BETTER_THAN
+                                            hmm_in_score(&nexthmm->hmm))) {
                                         hmm_enter(&nexthmm->hmm,
                                                   newscore,
                                                   hmm_out_history(&hmm->hmm),
@@ -500,8 +559,8 @@ fwdflat_prune_chan(ngram_search_t *ngs, int frame_idx)
                             /* Enter single word-internal phone. */
                             else {
                                 if ((hmm_frame(&nexthmm->hmm) < cf)
-                                    || (hmm_in_score(&nexthmm->hmm)
-                                        < newscore)) {
+                                    || (newscore BETTER_THAN
+                                        hmm_in_score(&nexthmm->hmm))) {
                                     hmm_enter(&nexthmm->hmm, newscore,
                                               hmm_out_history(&hmm->hmm), nf);
                                 }
@@ -510,7 +569,7 @@ fwdflat_prune_chan(ngram_search_t *ngs, int frame_idx)
                     }
                     /* Right-context phones - apply word beam and exit. */
                     else {
-                        if (newscore > wordthresh) {
+                        if (newscore BETTER_THAN wordthresh) {
                             ngram_search_save_bp(ngs, cf, w, newscore,
                                                  hmm_out_history(&hmm->hmm),
                                                  hmm->info.rc_id);
@@ -565,12 +624,12 @@ fwdflat_word_transition(ngram_search_t *ngs, int frame_idx)
     int32 cf, nf, b, thresh, pip, i, w, newscore;
     int32 best_silrc_score = 0, best_silrc_bp = 0;      /* FIXME: good defaults? */
     bptbl_t *bp;
-    dict_entry_t *de, *newde;
-    uint16 *rcpermtab;
     int32 *rcss;
     root_chan_t *rhmm;
     int32 *awl;
     float32 lwf;
+    dict_t *dict = ps_search_dict(ngs);
+    dict2pid_t *d2p = ps_search_dict2pid(ngs);
 
     cf = frame_idx;
     nf = cf + 1;
@@ -585,63 +644,84 @@ fwdflat_word_transition(ngram_search_t *ngs, int frame_idx)
 
     /* Scan words exited in current frame */
     for (b = ngs->bp_table_idx[cf]; b < ngs->bpidx; b++) {
+        xwdssid_t *rssid;
+        int32 silscore;
+
         bp = ngs->bp_table + b;
         ngs->word_lat_idx[bp->wid] = NO_BP;
 
         if (bp->wid == ps_search_finish_wid(ngs))
             continue;
 
-        de = ps_search_dict(ngs)->dict_list[bp->wid];
-        rcpermtab =
-            (bp->r_diph >=
-             0) ? ps_search_dict(ngs)->rcFwdPermTable[bp->r_diph] : ngs->zeroPermTab;
+        /* DICT2PID location */
+        /* Get the mapping from right context phone ID to index in the
+         * right context table and the bscore_stack. */
         rcss = ngs->bscore_stack + bp->s_idx;
+        if (bp->last2_phone == -1)
+            rssid = NULL;
+        else
+            rssid = dict2pid_rssid(d2p, bp->last_phone, bp->last2_phone);
 
         /* Transition to all successor words. */
         for (i = 0; ngs->expand_word_list[i] >= 0; i++) {
             int32 n_used;
+
             w = ngs->expand_word_list[i];
-            newde = ps_search_dict(ngs)->dict_list[w];
+
             /* Get the exit score we recorded in save_bwd_ptr(), or
              * something approximating it. */
-            newscore = rcss[rcpermtab[newde->ci_phone_ids[0]]];
+            if (rssid)
+                newscore = rcss[rssid->cimap[dict_first_phone(dict, w)]];
+            else
+                newscore = rcss[0];
+            if (newscore == WORST_SCORE)
+                continue;
             /* FIXME: Floating point... */
             newscore += lwf
-                * ngram_tg_score(ngs->lmset, newde->wid, bp->real_wid,
+                * ngram_tg_score(ngs->lmset,
+                                 dict_basewid(dict, w),
+                                 bp->real_wid,
                                  bp->prev_real_wid, &n_used);
             newscore += pip;
 
             /* Enter the next word */
-            if (newscore > thresh) {
+            if (newscore BETTER_THAN thresh) {
                 rhmm = (root_chan_t *) ngs->word_chan[w];
                 if ((hmm_frame(&rhmm->hmm) < cf)
-                    || (hmm_in_score(&rhmm->hmm) < newscore)) {
+                    || (newscore BETTER_THAN hmm_in_score(&rhmm->hmm))) {
                     hmm_enter(&rhmm->hmm, newscore, b, nf);
-                    if (hmm_is_mpx(&rhmm->hmm)) {
-                        rhmm->hmm.s.mpx_ssid[0] =
-                            ps_search_dict(ngs)->lcFwdTable[rhmm->diphone]
-                            [de->ci_phone_ids[de->len-1]];
-                    }
-
+                    /* DICT2PID: This is where mpx ssids get introduced. */
+                    /* Look up the ssid to use when entering this mpx triphone. */
+                    hmm_mpx_ssid(&rhmm->hmm, 0) =
+                        dict2pid_ldiph_lc(d2p, rhmm->ciphone, rhmm->ci2phone,
+                                          dict_last_phone(dict, bp->wid));
+                    assert(IS_S3SSID(hmm_mpx_ssid(&rhmm->hmm, 0)));
+                    E_DEBUG(6,("ssid %d(%d,%d) = %d\n",
+                               rhmm->ciphone, dict_last_phone(dict, bp->wid), rhmm->ci2phone,
+                               hmm_mpx_ssid(&rhmm->hmm, 0)));
                     bitvec_set(ngs->word_active, w);
                 }
             }
         }
 
         /* Get the best exit into silence. */
-        if (best_silrc_score < rcss[rcpermtab[ps_search_acmod(ngs)->mdef->sil]]) {
-            best_silrc_score = rcss[rcpermtab[ps_search_acmod(ngs)->mdef->sil]];
+        if (rssid)
+            silscore = rcss[rssid->cimap[ps_search_acmod(ngs)->mdef->sil]];
+        else
+            silscore = rcss[0];
+        if (silscore BETTER_THAN best_silrc_score) {
+            best_silrc_score = silscore;
             best_silrc_bp = b;
         }
     }
 
     /* Transition to <sil> */
     newscore = best_silrc_score + ngs->silpen + pip;
-    if ((newscore > thresh) && (newscore > WORST_SCORE)) {
+    if ((newscore BETTER_THAN thresh) && (newscore BETTER_THAN WORST_SCORE)) {
         w = ps_search_silence_wid(ngs);
         rhmm = (root_chan_t *) ngs->word_chan[w];
         if ((hmm_frame(&rhmm->hmm) < cf)
-            || (hmm_in_score(&rhmm->hmm) < newscore)) {
+            || (newscore BETTER_THAN hmm_in_score(&rhmm->hmm))) {
             hmm_enter(&rhmm->hmm, newscore,
                       best_silrc_bp, nf);
             bitvec_set(ngs->word_active, w);
@@ -649,14 +729,14 @@ fwdflat_word_transition(ngram_search_t *ngs, int frame_idx)
     }
     /* Transition to noise words */
     newscore = best_silrc_score + ngs->fillpen + pip;
-    if ((newscore > thresh) && (newscore > WORST_SCORE)) {
+    if ((newscore BETTER_THAN thresh) && (newscore BETTER_THAN WORST_SCORE)) {
         for (w = ps_search_silence_wid(ngs) + 1; w < ps_search_n_words(ngs); w++) {
             rhmm = (root_chan_t *) ngs->word_chan[w];
             /* Noise words that aren't a single phone will have NULL here. */
             if (rhmm == NULL)
                 continue;
             if ((hmm_frame(&rhmm->hmm) < cf)
-                || (hmm_in_score(&rhmm->hmm) < newscore)) {
+                || (newscore BETTER_THAN hmm_in_score(&rhmm->hmm))) {
                 hmm_enter(&rhmm->hmm, newscore,
                           best_silrc_bp, nf);
                 bitvec_set(ngs->word_active, w);
@@ -703,32 +783,29 @@ fwdflat_renormalize_scores(ngram_search_t *ngs, int frame_idx, int32 norm)
 }
 
 int
-ngram_fwdflat_search(ngram_search_t *ngs)
+ngram_fwdflat_search(ngram_search_t *ngs, int frame_idx)
 {
     int16 const *senscr;
-    int frame_idx, best_senid;
-    int16 best_senscr;
     int32 nf, i, j;
     int32 *nawl;
 
-    /* Determine if we actually have a frame to process. */
-    if (ps_search_acmod(ngs)->n_feat_frame == 0)
-        return 0;
-
     /* Activate our HMMs for the current frame if need be. */
     if (!ps_search_acmod(ngs)->compallsen)
-        compute_fwdflat_sen_active(ngs, acmod_frame_idx(ps_search_acmod(ngs)));
+        compute_fwdflat_sen_active(ngs, frame_idx);
 
     /* Compute GMM scores for the current frame. */
-    senscr = acmod_score(ps_search_acmod(ngs), &frame_idx,
-                         &best_senscr, &best_senid);
+    senscr = acmod_score(ps_search_acmod(ngs), &frame_idx);
     ngs->st.n_senone_active_utt += ps_search_acmod(ngs)->n_senone_active;
 
     /* Mark backpointer table for current frame. */
     ngram_search_mark_bptable(ngs, frame_idx);
 
-    /* Renormalize if necessary (FIXME: Make sure to test this) */
-    if (ngs->best_score + (2 * ngs->beam) < WORST_SCORE) {
+    /* If the best score is equal to or worse than WORST_SCORE,
+     * recognition has failed, don't bother to keep trying. */
+    if (ngs->best_score == WORST_SCORE || ngs->best_score WORSE_THAN WORST_SCORE)
+        return 0;
+    /* Renormalize if necessary */
+    if (ngs->best_score + (2 * ngs->beam) WORSE_THAN WORST_SCORE) {
         E_INFO("Renormalizing Scores at frame %d, best score %d\n",
                frame_idx, ngs->best_score);
         fwdflat_renormalize_scores(ngs, frame_idx, ngs->best_score);
@@ -794,18 +871,13 @@ static void
 destroy_fwdflat_chan(ngram_search_t *ngs)
 {
     int32 i, wid;
-    dict_entry_t *de;
 
     for (i = 0; ngs->fwdflat_wordlist[i] >= 0; i++) {
         root_chan_t *rhmm;
         chan_t *thmm;
         wid = ngs->fwdflat_wordlist[i];
-        de = ps_search_dict(ngs)->dict_list[wid];
-
-        if (de->len == 1)
+        if (dict_is_single_phone(ps_search_dict(ngs),wid))
             continue;
-
-        assert(de->mpx);
         assert(ngs->word_chan[wid] != NULL);
 
         /* The first HMM in ngs->word_chan[wid] was allocated with
@@ -831,7 +903,7 @@ ngram_fwdflat_finish(ngram_search_t *ngs)
     bitvec_clear_all(ngs->word_active, ps_search_n_words(ngs));
 
     /* This is the number of frames processed. */
-    cf = acmod_frame_idx(ps_search_acmod(ngs));
+    cf = ps_search_acmod(ngs)->output_frame;
     /* Add a mark in the backpointer table for one past the final frame. */
     ngram_search_mark_bptable(ngs, cf);
 
