@@ -48,12 +48,14 @@
 #include <ckd_alloc.h>
 #include <listelem_alloc.h>
 #include <strfuncs.h>
+#include <err.h>
 #include <pio.h>
 
 /* Local headers. */
 #include "pocketsphinx_internal.h"
 #include "ps_lattice_internal.h"
 #include "ngram_search.h"
+#include "dict.h"
 
 /*
  * Create a directed link between "from" and "to" nodes, but if a link already exists,
@@ -92,7 +94,7 @@ ps_lattice_link(ps_lattice_t *dag, ps_latnode_t *from, ps_latnode_t *to, int32 s
     }
     else {
         /* Link already exists; just retain the best ascr */
-        if (fwdlink->link->ascr < score) {
+        if (score BETTER_THAN fwdlink->link->ascr) {
             fwdlink->link->ascr = score;
             fwdlink->link->ef = ef;
         }
@@ -108,7 +110,7 @@ ps_lattice_bypass_fillers(ps_lattice_t *dag, int32 silpen, int32 fillpen)
     /* Bypass filler nodes */
     for (node = dag->nodes; node; node = node->next) {
         latlink_list_t *revlink;
-        if (node == dag->end || !ISA_FILLER_WORD(dag->search, node->basewid))
+        if (node == dag->end || !dict_filler_word(ps_search_dict(dag->search), node->basewid))
             continue;
 
         /* Replace each link entering filler node with links to all its successors */
@@ -125,7 +127,7 @@ ps_lattice_bypass_fillers(ps_lattice_t *dag, int32 silpen, int32 fillpen)
              */
             for (forlink = node->exits; forlink; forlink = forlink->next) {
                 ps_latlink_t *flink = forlink->link;
-                if (!ISA_FILLER_WORD(dag->search, flink->to->basewid)) {
+                if (!dict_filler_word(ps_search_dict(dag->search), flink->to->basewid)) {
                     ps_lattice_link(dag, rlink->from, flink->to,
                                     score + flink->ascr, flink->ef);
                 }
@@ -254,7 +256,7 @@ ps_lattice_write(ps_lattice_t *dag, char const *filename)
     for (i = 0, d = dag->nodes; d; d = d->next, i++) {
         d->id = i;
         fprintf(fp, "%d %s %d %d %d\n",
-                i, dict_word_str(ps_search_dict(dag->search), d->wid),
+                i, dict_wordstr(ps_search_dict(dag->search), d->wid),
                 d->sf, d->fef, d->lef);
     }
     fprintf(fp, "#\n");
@@ -412,7 +414,7 @@ ps_lattice_read(ps_decoder_t *ps,
             goto load_error;
         }
 
-        w = dict_to_id(ps->dict, wd);
+        w = dict_wordid(ps->dict, wd);
         if (w < 0) {
             E_ERROR("Unknown word in line: %s\n", line->buf);
             goto load_error;
@@ -426,7 +428,7 @@ ps_lattice_read(ps_decoder_t *ps,
         d = listelem_malloc(dag->latnode_alloc);
         darray[i] = d;
         d->wid = w;
-        d->basewid = dict_base_wid(ps->dict, w);
+        d->basewid = dict_basewid(ps->dict, w);
         d->id = seqid;
         d->sf = sf;
         d->fef = fef;
@@ -505,7 +507,7 @@ ps_lattice_read(ps_decoder_t *ps,
     /* Minor hack: If the final node is a filler word and not </s>,
      * then set its base word ID to </s>, so that the language model
      * scores won't be screwed up. */
-    if (ISA_FILLER_WORD(dag->search, dag->end->wid))
+    if (dict_filler_word(ps_search_dict(dag->search), dag->end->wid))
         dag->end->basewid = ps_search_finish_wid(dag->search);
 
     /* Mark reachable from dag->end */
@@ -619,13 +621,13 @@ ps_latnode_times(ps_latnode_t *node, int16 *out_fef, int16 *out_lef)
 char const *
 ps_latnode_word(ps_lattice_t *dag, ps_latnode_t *node)
 {
-    return dict_word_str(ps_search_dict(dag->search), node->wid);
+    return dict_wordstr(ps_search_dict(dag->search), node->wid);
 }
 
 char const *
 ps_latnode_baseword(ps_lattice_t *dag, ps_latnode_t *node)
 {
-    return dict_word_str(ps_search_dict(dag->search), node->basewid);
+    return dict_wordstr(ps_search_dict(dag->search), node->basewid);
 }
 
 int32
@@ -701,7 +703,7 @@ ps_latlink_word(ps_lattice_t *dag, ps_latlink_t *link)
 {
     if (link->from == NULL)
         return NULL;
-    return dict_word_str(ps_search_dict(dag->search), link->from->wid);
+    return dict_wordstr(ps_search_dict(dag->search), link->from->wid);
 }
 
 char const *
@@ -709,7 +711,7 @@ ps_latlink_baseword(ps_lattice_t *dag, ps_latlink_t *link)
 {
     if (link->from == NULL)
         return NULL;
-    return dict_word_str(ps_search_dict(dag->search), link->from->basewid);
+    return dict_wordstr(ps_search_dict(dag->search), link->from->basewid);
 }
 
 ps_latlink_t *
@@ -735,31 +737,32 @@ ps_lattice_hyp(ps_lattice_t *dag, ps_latlink_t *link)
 
     /* Backtrace once to get hypothesis length. */
     len = 0;
-    if (ISA_REAL_WORD(dag->search, link->to->basewid))
-        len += strlen(dict_word_str(ps_search_dict(dag->search), link->to->basewid)) + 1;
+    /* FIXME: There may not be a search, but actually there should be a dict. */
+    if (dict_real_word(ps_search_dict(dag->search), link->to->basewid))
+        len += strlen(dict_wordstr(ps_search_dict(dag->search), link->to->basewid)) + 1;
     for (l = link; l; l = l->best_prev) {
-        if (ISA_REAL_WORD(dag->search, l->from->basewid))
-            len += strlen(dict_word_str(ps_search_dict(dag->search), l->from->basewid)) + 1;
+        if (dict_real_word(ps_search_dict(dag->search), l->from->basewid))
+            len += strlen(dict_wordstr(ps_search_dict(dag->search), l->from->basewid)) + 1;
     }
 
     /* Backtrace again to construct hypothesis string. */
     ckd_free(dag->hyp_str);
-    dag->hyp_str = ckd_calloc(1, len);
+    dag->hyp_str = ckd_calloc(1, len+1); /* extra one incase the hyp is empty */
     c = dag->hyp_str + len - 1;
-    if (ISA_REAL_WORD(dag->search, link->to->basewid)) {
-        len = strlen(dict_word_str(ps_search_dict(dag->search), link->to->basewid));
+    if (dict_real_word(ps_search_dict(dag->search), link->to->basewid)) {
+        len = strlen(dict_wordstr(ps_search_dict(dag->search), link->to->basewid));
         c -= len;
-        memcpy(c, dict_word_str(ps_search_dict(dag->search), link->to->basewid), len);
+        memcpy(c, dict_wordstr(ps_search_dict(dag->search), link->to->basewid), len);
         if (c > dag->hyp_str) {
             --c;
             *c = ' ';
         }
     }
     for (l = link; l; l = l->best_prev) {
-        if (ISA_REAL_WORD(dag->search, l->from->basewid)) {
-            len = strlen(dict_word_str(ps_search_dict(dag->search), l->from->basewid));
+        if (dict_real_word(ps_search_dict(dag->search), l->from->basewid)) {
+            len = strlen(dict_wordstr(ps_search_dict(dag->search), l->from->basewid));
             c -= len;
-            memcpy(c, dict_word_str(ps_search_dict(dag->search), l->from->basewid), len);
+            memcpy(c, dict_wordstr(ps_search_dict(dag->search), l->from->basewid), len);
             if (c > dag->hyp_str) {
                 --c;
                 *c = ' ';
@@ -846,7 +849,7 @@ ps_lattice_link2itor(ps_seg_t *seg, ps_latlink_t *link, int to)
             }
         }
     }
-    seg->word = dict_word_str(ps_search_dict(seg->search), node->wid);
+    seg->word = dict_wordstr(ps_search_dict(seg->search), node->wid);
     seg->sf = node->sf;
     seg->ascr = link->ascr;
     /* Compute language model score from best predecessors. */
@@ -1125,7 +1128,7 @@ ps_lattice_bestpath(ps_lattice_t *dag, ngram_model_t *lmset,
         int32 n_used;
 
         /* Ignore filler words. */
-        if (ISA_FILLER_WORD(search, x->link->to->basewid)
+        if (dict_filler_word(ps_search_dict(search), x->link->to->basewid)
             && x->link->to != dag->end)
             continue;
 
@@ -1147,9 +1150,9 @@ ps_lattice_bestpath(ps_lattice_t *dag, ngram_model_t *lmset,
         int32 bprob, n_used;
 
         /* Skip filler nodes in traversal. */
-        if (ISA_FILLER_WORD(search, link->from->basewid) && link->from != dag->start)
+        if (dict_filler_word(ps_search_dict(search), link->from->basewid) && link->from != dag->start)
             continue;
-        if (ISA_FILLER_WORD(search, link->to->basewid) && link->to != dag->end)
+        if (dict_filler_word(ps_search_dict(search), link->to->basewid) && link->to != dag->end)
             continue;
 
         /* Sanity check, we should not be traversing edges that
@@ -1172,7 +1175,7 @@ ps_lattice_bestpath(ps_lattice_t *dag, ngram_model_t *lmset,
             int32 tscore, score;
 
             /* Skip links to filler words in update. */
-            if (ISA_FILLER_WORD(search, x->link->to->basewid)
+            if (dict_filler_word(ps_search_dict(search), x->link->to->basewid)
                 && x->link->to != dag->end)
                 continue;
 
@@ -1187,7 +1190,7 @@ ps_lattice_bestpath(ps_lattice_t *dag, ngram_model_t *lmset,
                 tscore = 0;
             /* Update link score with maximum link score. */
             score = link->path_scr + tscore + x->link->ascr;
-            if (score > x->link->path_scr) {
+            if (score BETTER_THAN x->link->path_scr) {
                 x->link->path_scr = score;
                 x->link->best_prev = link;
             }
@@ -1205,7 +1208,7 @@ ps_lattice_bestpath(ps_lattice_t *dag, ngram_model_t *lmset,
     for (x = dag->end->entries; x; x = x->next) {
         int32 bprob, n_used;
 
-        if (ISA_FILLER_WORD(search, x->link->from->basewid))
+        if (dict_filler_word(ps_search_dict(search), x->link->from->basewid))
             continue;
         if (lmset)
             bprob = ngram_ng_prob(lmset,
@@ -1214,7 +1217,7 @@ ps_lattice_bestpath(ps_lattice_t *dag, ngram_model_t *lmset,
         else
             bprob = 0;
         dag->norm = logmath_add(lmath, dag->norm, x->link->alpha + bprob);
-        if (x->link->path_scr > bestescr) {
+        if (x->link->path_scr BETTER_THAN bestescr) {
             bestescr = x->link->path_scr;
             bestend = x->link;
         }
@@ -1223,7 +1226,7 @@ ps_lattice_bestpath(ps_lattice_t *dag, ngram_model_t *lmset,
     dag->norm += (int32)dag->final_node_ascr * ascale;
 
     E_INFO("Normalizer P(O) = alpha(%s:%d:%d) = %d\n",
-           dict_word_str(dag->search->dict, dag->end->wid),
+           dict_wordstr(dag->search->dict, dag->end->wid),
            dag->end->sf, dag->end->lef,
            dag->norm);
     return bestend;
@@ -1294,9 +1297,9 @@ ps_lattice_posterior(ps_lattice_t *dag, ngram_model_t *lmset,
         int32 bprob, n_used;
 
         /* Skip filler nodes in traversal. */
-        if (ISA_FILLER_WORD(search, link->from->basewid) && link->from != dag->start)
+        if (dict_filler_word(ps_search_dict(search), link->from->basewid) && link->from != dag->start)
             continue;
-        if (ISA_FILLER_WORD(search, link->to->basewid) && link->to != dag->end)
+        if (dict_filler_word(ps_search_dict(search), link->to->basewid) && link->to != dag->end)
             continue;
 
         /* Calculate LM probability. */
@@ -1310,7 +1313,7 @@ ps_lattice_posterior(ps_lattice_t *dag, ngram_model_t *lmset,
             /* Track the best path - we will backtrace in order to
                calculate the unscaled joint probability for sentence
                posterior. */
-            if (link->path_scr > bestescr) {
+            if (link->path_scr BETTER_THAN bestescr) {
                 bestescr = link->path_scr;
                 bestend = link;
             }
@@ -1320,7 +1323,7 @@ ps_lattice_posterior(ps_lattice_t *dag, ngram_model_t *lmset,
         else {
             /* Update beta from all outgoing betas. */
             for (x = link->to->exits; x; x = x->next) {
-                if (ISA_FILLER_WORD(search, x->link->to->basewid) && x->link->to != dag->end)
+                if (dict_filler_word(ps_search_dict(search), x->link->to->basewid) && x->link->to != dag->end)
                     continue;
                 link->beta = logmath_add(lmath, link->beta,
                                          x->link->beta + bprob + x->link->ascr * ascale);
@@ -1364,7 +1367,7 @@ best_rem_score(ps_astar_t *nbest, ps_latnode_t * from)
         if (nbest->lmset)
             score += ngram_bg_score(nbest->lmset, x->link->to->basewid,
                                     from->basewid, &n_used) * nbest->lwf;
-        if (score > bestscore)
+        if (score BETTER_THAN bestscore)
             bestscore = score;
     }
     from->info.rem_score = bestscore;
@@ -1496,7 +1499,7 @@ ps_astar_start(ps_lattice_t *dag,
     nbest->lwf = lwf;
     nbest->sf = sf;
     if (ef < 0)
-        nbest->ef = dag->n_frames - ef;
+        nbest->ef = dag->n_frames + 1;
     else
         nbest->ef = ef;
     nbest->w1 = w1;
@@ -1541,36 +1544,30 @@ ps_astar_start(ps_lattice_t *dag,
 ps_latpath_t *
 ps_astar_next(ps_astar_t *nbest)
 {
-    ps_latpath_t *top;
     ps_lattice_t *dag;
 
     dag = nbest->dag;
 
     /* Pop the top (best) partial hypothesis */
-    while ((top = nbest->path_list) != NULL) {
+    while ((nbest->top = nbest->path_list) != NULL) {
         nbest->path_list = nbest->path_list->next;
-        if (top == nbest->path_tail)
+        if (nbest->top == nbest->path_tail)
             nbest->path_tail = NULL;
         nbest->n_path--;
 
         /* Complete hypothesis? */
-        if ((top->node->sf >= nbest->ef)
-            || ((top->node == dag->end) &&
+        if ((nbest->top->node->sf >= nbest->ef)
+            || ((nbest->top->node == dag->end) &&
                 (nbest->ef > dag->end->sf))) {
-            /* FIXME: Verify that it is non-empty. */
-            return top;
+            /* FIXME: Verify that it is non-empty.  Also we may want
+             * to verify that it is actually distinct from other
+             * paths, since often this is not the case*/
+            return nbest->top;
         }
         else {
-            if (top->node->fef < nbest->ef)
-                path_extend(nbest, top);
+            if (nbest->top->node->fef < nbest->ef)
+                path_extend(nbest, nbest->top);
         }
-
-        /*
-         * Add top to paths already processed; cannot be freed because other paths
-         * point to it.
-         */
-        top->next = nbest->paths_done;
-        nbest->paths_done = top;
     }
 
     /* Did not find any more paths to extend. */
@@ -1591,18 +1588,22 @@ ps_astar_hyp(ps_astar_t *nbest, ps_latpath_t *path)
     /* Backtrace once to get hypothesis length. */
     len = 0;
     for (p = path; p; p = p->parent) {
-        if (ISA_REAL_WORD(search, p->node->basewid))
-            len += strlen(dict_word_str(ps_search_dict(search), p->node->basewid)) + 1;
+        if (dict_real_word(ps_search_dict(search), p->node->basewid))
+            len += strlen(dict_wordstr(ps_search_dict(search), p->node->basewid)) + 1;
+    }
+
+    if (len == 0) {
+	return NULL;
     }
 
     /* Backtrace again to construct hypothesis string. */
     hyp = ckd_calloc(1, len);
     c = hyp + len - 1;
     for (p = path; p; p = p->parent) {
-        if (ISA_REAL_WORD(search, p->node->basewid)) {
-            len = strlen(dict_word_str(ps_search_dict(search), p->node->basewid));
+        if (dict_real_word(ps_search_dict(search), p->node->basewid)) {
+            len = strlen(dict_wordstr(ps_search_dict(search), p->node->basewid));
             c -= len;
-            memcpy(c, dict_word_str(ps_search_dict(search), p->node->basewid), len);
+            memcpy(c, dict_wordstr(ps_search_dict(search), p->node->basewid), len);
             if (c > hyp) {
                 --c;
                 *c = ' ';
@@ -1626,7 +1627,7 @@ ps_astar_node2itor(astar_seg_t *itor)
         seg->ef = node->lef;
     else
         seg->ef = itor->nodes[itor->cur + 1]->sf - 1;
-    seg->word = dict_word_str(ps_search_dict(seg->search), node->wid);
+    seg->word = dict_wordstr(ps_search_dict(seg->search), node->wid);
     seg->sf = node->sf;
     seg->prob = 0; /* FIXME: implement forward-backward */
 }
