@@ -13,13 +13,13 @@ main(int argc, char *argv[])
 
 	TEST_ASSERT(config =
 		    cmd_ln_init(NULL, ps_args(), TRUE,
-				"-hmm", MODELDIR "/hmm/wsj1",
-				"-lm", MODELDIR "/lm/wsj/wlist5o.3e-7.vp.tg.lm.DMP",
-				"-dict", MODELDIR "/lm/cmudict.0.6d",
+				"-hmm", MODELDIR "/hmm/en_US/hub4wsj_sc_8k",
+				"-lm", MODELDIR "/lm/en_US/wsj0vp.5000.DMP",
+				"-dict", MODELDIR "/lm/en_US/cmu07a.dic",
 				"-fwdtree", "no",
 				"-fwdflat", "yes",
 				"-bestpath", "yes",
 				"-input_endian", "little",
 				"-samprate", "16000", NULL));
-	return ps_decoder_test(config, "BESTPATH", "GO FORWARD TEN YEARS");
+	return ps_decoder_test(config, "BESTPATH", "go forward and users");
 }
